with Strings, Globals, ScalVectors.Eigenv, ScalVectors.Eigenv.RNum;
use Strings, Globals;

pragma Elaborate_All (Strings,Globals,ScalVectors.Eigenv.RNum);

package body Cylinders.Fix is

  use Lin,Lin_Contr;

  procedure Lin1(F: in out Cylinder; LF: in out Cylinder) is
  begin
    Mult(SmallCosX,F);
    Lq(Two,F,LF);
  end Lin1;

  procedure Lin2(F: in Cylinder; LF: in out Cylinder) is
    --- (1 - 2r cos(nx) e^(-ny) + r^2 e^(-2ny)) e^(-2y)
    --- r^2 is assumed to be representable as Flt
    S: FScalar;
    G: Cylinder;
  begin
    Lq(Two,F,LF);            ---  LF := e^(-2y) F

    Lq(Flt(NN+2),F,G);       ---  e^(-(2+n)y) F
    Prod(Two*RR,CosNX,S);
    Mult(S,G);               ---  G = (2r cos(nx)) e^(-(2+n)y) F
    Sub(G,LF);               ---  LF := LF-G

    Lq(Flt(2*NN+2),F,G);     ---  e^(-(2n+2)y) F
    Mult(RR*RR,G);           ---  G = r^2 e^(-(2n+2)y) F3
    Add(G,LF);               ---  LF := LF+G
  end Lin2;

  procedure NegLapGMap(F: in Cylinder; GF: in out Cylinder) is
    --- GMap without NegInvLap
    --- can be used e.g. to compute gradient
    FFF: Cylinder;
  begin
    Prod(F,F,FFF);
    Mult(F,FFF);     ---  F^3
    case Map is
      when 1 => Lq(Two,FFF,GF);
      when 2 => Lin2(FFF,GF);
      when 3 => Apply(LT,FFF,GF);
      when 4 => Apply(LT,FFF,GF);
      when 5 => Apply(LT,FFF,GF);
      when 6 => Apply(LT,FFF,GF);
      when others => raise Not_Implemented;
    end case;
  end NegLapGMap;

  procedure GMap(F: in Cylinder; GF: in out Cylinder) is
  begin
    NegLapGMap(F,GF);
    NegInvLap(GF);
  end GMap;

  procedure DGMap(F,FF3,H: in Cylinder; DGH: in out Cylinder) is
    pragma Unreferenced(F);
    --- FF3 = 3*F*F
    HFF3: Cylinder;
  begin
    Prod(H,FF3,HFF3);
    case Map is
      when 1 => Lq(Two,HFF3,DGH);
      when 2 => Lin2(HFF3,DGH);
      when 3 => Apply(LT,HFF3,DGH);
      when 4 => Apply(LT,HFF3,DGH);
      when 5 => Apply(LT,HFF3,DGH);
      when 6 => Apply(LT,HFF3,DGH);
      when others => raise Not_Implemented;
    end case;
    NegInvLap(DGH);
  end DGMap;

  procedure DContrNorm(M: in SMatrix; F: in Cylinder; PackSize: in Positive; R: out Radius) is
    Q: constant CylinderModes := Make(RadX,PackSize,True,False);  --- Err modes, but Extract is not used
    FF3: Cylinder;
    procedure DG(H: in Cylinder; DGH: in out Cylinder) is begin DGMap(F,FF3,H,DGH); end DG;
    procedure DContract is new DContr (DMap => DG);
    procedure DC(H: in Cylinder; DCH: in out Cylinder) is begin DContract(QC,M,H,DCH); end DC;
    procedure DCNorm is new Op_Norm (LinOp => DC);
  begin
    if PackSize <= Mat_PackSize then raise Not_Implemented; end if;
    Prod(F,F,FF3);
    Mult(Three,FF3);
    if Verbosity>1 then
      Show1("DContrNorm: Q'Last = ",Q'Last);
    end if;
    DCNorm(Q,R);
  end DContrNorm;

  procedure ContrFix(M: in SMatrix; PackSize: in Positive; F: in out Cylinder) is
    KMax: constant Radius := Flt(7)/Flt(8);
    E,K: Radius;
    ANorm,R,Eps,Del: Scalar;
    F0: Cylinder;
  begin
    Show0("ContrFix: ",False);
    Copy(F,F0);
    Center(F0);
    GMap(F0,F);
    Sub(F0,F);
    E := MaxNorm(F);
    Assign(E,Eps);
    OpNorm(QC,M,ANorm);
    Add(SOne,ANorm);
    Show2("Eps,ANorm ",E,Sup(ANorm));
    ---
    Del := Eps/(SOne-Scal(KMax))+Scal(Tiny);
    R := ANorm*Del;
    Copy(F0,F);
    AddBall(MaxNorm(R),F);
    DContrNorm(M,F,PackSize,K);
    Show1("ContrFix: K ",K);
    if K>KMax then raise Sorry with "K too large"; end if;
    ---
    Del := Eps/(SOne-Scal(K))+Scal(Tiny);
    R := ANorm*Del;
    Show2("ContrFix: Del,R: ",Sup(Del),Sup(R));
    Copy(F0,F);
    AddBall(MaxNorm(R),F);
  end ContrFix;

  --------------------------------------------
  --------------------------------------------

  procedure DerMatrix(F: in Cylinder; M: in out SMatrix) is
    --- for numeric mode only
    FF3: Cylinder;
    procedure DG(H: in Cylinder; DGH: in out Cylinder) is begin DGMap(F,FF3,H,DGH); end DG;
    procedure Enclose is new Enclosure (LinOp => DG);
  begin
    if Not_STrunc then raise Sorry; end if;
    Prod(F,F,FF3);
    Mult(Three,FF3);
    Enclose(QC,M);
  end DerMatrix;

  procedure ContrMatrix(F: in Cylinder; M: in out SMatrix) is
  begin
    if Verbosity>1 then Show0("running ContrMatrix"); end if;
    DerMatrix(F,M);
    Contr_Matrix(M);
  end ContrMatrix;

  procedure ApproxFix(F: in out Cylinder; M: in out SMatrix; Eps: in Radius := Zero) is
    --- find approximate fixed point
    procedure Contract0 is new ContrTmp0 (Map => GMap);
    Iter: constant Integer := 1024;
    MaxCredit: constant Integer := 5;
    Credit: Integer := MaxCredit;
    Err,MinErr: Radius := 1.0E+9;
    ErrMat: constant Radius := 1.0e-4;
    G,Tmp1,Tmp2: Cylinder;
  begin
    if Not_STrunc then raise Sorry with "only for numeric mode"; end if;
    if Verbosity>1 then Show0("running ApproxFix"); end if;
    Copy(F,G);
    for I in 1 .. Iter loop
      if Err>ErrMat then ContrMatrix(G,M); end if;
      Contract0(QC,M,G,Tmp1,Tmp2,Err);
      if Verbosity>1 then Show1("ApproxFix: step " & Strng(I) & ", Err ",Err); end if;
      if Err<MinErr then
        Copy(G,F);
        MinErr := Err;
        if (Err<Eps) or (Err=MinErr) then exit; end if;
        Credit := MaxCredit;
      else
        exit when Credit=0;
        Credit := Credit-1;
      end if;
    end loop;
    Show1("ApproxFix: Err ",MinErr);
    --    ContrMatrix(F,M);
  end ApproxFix;

  procedure IterContr(F: in out Cylinder; M: in out SMatrix) is
    --- improve approximate fixed point
    procedure Contract0 is new ContrTmp0 (Map => GMap);
    MaxCredit: constant Integer := 16;
    Credit: Integer := MaxCredit;
    Err,MinErr: Radius := One;
    G,Tmp1,Tmp2: Cylinder;
  begin
    if Not_STrunc then raise Sorry with "only for numeric mode"; end if;
    if Verbosity>1 then Show0("running IterContr"); end if;
    Copy(F,G);
    --ContrMatrix(G,M);
    while Credit>0 loop
      Contract0(QC,M,G,Tmp1,Tmp2,Err);
      Show1("IterContr: Err ",Err);
      if Err<MinErr then
        Copy(G,F);
        MinErr := Err;
        Credit := MaxCredit;
        --ContrMatrix(G,M);
      else
        Credit := Credit-1;
      end if;
    end loop;
    Show1("IterContr: Err ",MinErr);
    --ContrMatrix(F,M);
  end IterContr;

  procedure EigenDer(Name: in String; F: in Cylinder) is
    A,Vr,Vi: QC_Matrix;
    Wr,Wi: QC_Vector;
    FHat: Cylinder;
    package SVE is new SSV.Eigenv;
    package SVEN is new SVE.RNum;
    use SV,SSV,SVEN;

    procedure AddProd(V: in QC_Vector) is
    begin
      for I in V'Range loop
        AddProd(QC(I),V(I),FHat,False);
      end loop;
    end AddProd;

  begin
    DerMatrix(F,A);
    Eigen(A,Wr,Wi,Vr,Vi);
    SetZero(1,FHat);
    for I in 1..50 loop
      if Approx(Wr(I))>0.5 then
        Show2("ev:",Wr(I),Wi(I));
      end if;
      if abs(Approx(Wr(I))-One)<0.01 then
        AddProd(GetColumn(I,Vr));
      end if;
    end loop;
    --    Write(Name,FHat);
  end EigenDer;

  ----------------- plotting

  procedure PlotWPhi(FileName: in String; XL,YL: in Positive := 64) is
    --- plot |Phi'|^2
    DX: constant Flt := Two*Approx_Pi/Flt(XL);
    DY: constant Flt := EllR/Flt(YL);
    X,Y,Z: Flt;
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Writing " & FileName); end if;
    Create(F,Out_File,FileName);
    for M in 0 .. XL loop
      X := Flt(M)*DX;
      for N in 0 .. YL loop
        Y := Flt(N)*DY;
        case Map is
          when 1 => Z := WPhi1(Y);
          when 2 => Z := WPhi2(X,Y,RR,NN);
          when 3 => Z := WPhi3(X,Y,RR);
          when 4 => Z := WPhi4(X,Y,RR);
          when 5 => Z := WPhi5(X,Y,RR);
          when 6 => Z := WPhi4(X,Y,RR);
          when others => raise Not_Implemented;
        end case;
        Plot((X,Y,Z),F);
      end loop;
      if Plot_Newline then New_Line(F); end if;
    end loop;
    Close(F);
  end PlotWPhi;

  procedure PlotJDensity(C: in Cylinder; FileName: in String; XL,YL: in Positive := 64) is
    --- plot |gradient A|^2(x,y) - (1/4)|Phi'|^2 A(x,y)^4
    DX: constant Flt := Two*Approx_Pi/Flt(XL);
    DY: constant Flt := EllR/Flt(YL);
    X,Y,Z,W,XFac,YFac: Flt;
    Integral: Flt := Zero;
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Writing " & FileName); end if;
    Create(F,Out_File,FileName);
    for M in 0 .. XL loop
      if M=0 or else M=XL then XFac := Half*DX; else XFac := DX; end if;
      X := Flt(M)*DX;
      for N in 0 .. YL loop
        if N=0 or else N=YL then YFac := Half*DY; else YFac := DY; end if;
        Y := Flt(N)*DY;
        case Map is
          when 1 => W := WPhi1(Y);
          when 2 => W := WPhi2(X,Y,RR,NN);
          when 3 => W := WPhi3(X,Y,RR);
          when 4 => W := WPhi4(X,Y,RR);
          when 5 => W := WPhi5(X,Y,RR);
          when 6 => W := WPhi4(X,Y,RR);
          when others => raise Not_Implemented;
        end case;
        Z := ApproxGrad2(X,Y,C)-Quarter*W*ApproxVal(X,Y,C)**4;
        Plot((X,Y,Z),F);
        Integral := Integral+Z*XFac*YFac; --- Simpson's rule
      end loop;
      if Plot_Newline then New_Line(F); end if;
    end loop;
    Close(F);
    if Verbosity>0 then Show1("J =",Integral); end if;
  end PlotJDensity;

  procedure Plot3(C: in Cylinder; FileName: in String; XL,YL: in Positive := 64) is
    DX: constant Flt := Two*Approx_Pi/Flt(XL);
    DY: constant Flt := EllR/Flt(YL);
    P,Q: Point;
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Writing " & FileName); end if;
    Create(F,Out_File,FileName);
    for M in 0 .. XL loop
      P.X := Flt(M)*DX;
      for N in 0 .. YL loop
        P.Y := Flt(N)*DY;
        P.Z := ApproxVal(P.X,P.Y,C);
        case Map is
          when 1 => Q := Phi1(P);
          when 2 => Q := Phi2(P,RR,NN);
          when 3 => Q := Phi3(P,RR);
          when 4 => Q := Phi4(P,RR);
          when 5 => Q := Phi5(P,RR);
          when 6 => Q := Phi6(P,RR);
          when others => raise Not_Implemented;
        end case;
        Plot(Q,F);
      end loop;
      if Plot_Newline then New_Line(F); end if;
    end loop;
    Close(F);
  end Plot3;

begin

  case Map is
    when 3 => Disk_Iso_Plus(RR,LT);
    when 4 => Cassini_Plus(RR,LT);
    when 5 => InvEllipse_Plus(RR,LT);
    when 6 => Near_Cassini_Plus(RR,LT,NN);
    when others => null;
  end case;

end Cylinders.Fix;
