with Strings, Globals, Protected_Counters, Flts.Complexs;
use Strings, Globals, Protected_Counters;

pragma Elaborate_All (Strings,Globals,Protected_Counters,Flts.Complexs);

package body Cylinders is

  use LTays;

  procedure Show_Cy_Param(Full: in Boolean := True) is
    use Flt_EF;
  begin
    Show1("Cy:          EllR = ",EllR);
    Show1("Cy:   radii ratio = ",Exp(-EllR));
    Show_F1_Param(False);
    Show_Cp_Param(Full);
  end Show_Cy_Param;

  function Read(FileName: String; Decimal: Boolean := False) return FltVec is
    L: Integer;
    F: File_Type;
  begin
    Show0("Reading " & FileName);
    Open(F,In_File,FileName);
    Int_IO.Get(F,L);
    declare
      V: FltVec(1 .. L);
    begin
      for I in 1 .. L loop
        Get(F,V(I),Decimal);
      end loop;
      Close(F);
      return V;
    end;
  end Read;

  procedure Mult(S: in Scalar; A: in out Cylinder) is
    C: FPolyY renames A.C;
  begin
    for I in 0 .. DegY loop Mult(S,C(I)); end loop;
    if Not_STrunc then
      declare
        E: ErrsY renames A.E;
        R: constant Radius := MaxNorm(S);
      begin
        for I in EFreqY loop E(I) := R*E(I); end loop;
      end;
    end if;
  end Mult;

  function Sigma(QDivKap2: Scalar) return Sums is
    -- QDivKap2 = (q/kappa)^2
    RNegQDivKap2: constant Flt := -Inf(QDivKap2);
    SigmaK: Flt := Two/Flt(2*KMax+1);
    S: Sums;
  begin
    for K in reverse 0 .. KMax loop
      SigmaK := SigmaK+NegOne/(RNegQDivKap2-Flt(K*K)); -- add 1/(k^2+(q/kappa)^2)
      if K<=LSums then S(K) := SigmaK; end if;
    end loop;
    return S;
  end Sigma;

  procedure Init(Q: in Flt; O: in out QData) is
    Eql: Scalar renames O.FourK0;
    Tmp: Scalar renames O.FourK1;
  begin
    Prod(Q,InvKappa,O.QDivKap);           -- QDivKap = q/kappa
    Prod(O.QDivKap,O.QDivKap,O.QDivKap2); -- QDivKap2 = (q/kappa)^2
    Prod(-Q,Ell,Tmp);
    Exp(Tmp,Eql);                   -- Eql =  e^{-ql}
    Diff(SOne,Eql,O.K0);            -- K0 = 1-e^{-ql}
    Sum(SOne,Eql,O.K1);             -- K1 = 1+e^{-ql}
    Prod(Q,InvEllKap2,Tmp);         --                 q/(l*kappa^2)
    Mult(Tmp,O.K0);                 -- K0 = (1-e^{-ql})q/(l*kappa^2)
    Mult(Tmp,O.K1);                 -- K1 = (1+e^{-ql})q/(l*kappa^2)
                                    --
    Prod(STwoOrFour,O.K0,O.FourK0); -- FourK0 = 4*K0 (or 2*K0 if TinyWFac)
    Prod(STwoOrFour,O.K1,O.FourK1); -- FourK1 = 4*K1 (or 2*K1 if TinyWFac)
    if Not_STrunc then
      O.Sigma := Sigma(O.QDivKap2); -- Sigma(t) = Sum_{k>=t} 1/(k^2+(q/kappa)^2)
    end if;
  end Init;

  procedure Lqjn(O: in QData; J,N: in Positive; S: in out FScalar) is
    --- for loops
    NmJ: constant Integer := N-J;
    NpJ: constant Integer := N+J;
    Num,Den: FScalar;
    T: FScalar renames Num;
  begin
    Assign(NmJ*NmJ,S);
    Add(O.QDivKap2,S);      -- S = (n-j)^2+(q/kappa)^2
    Assign(NpJ*NpJ,T);
    Add(O.QDivKap2,T);      -- T = (n+j)^2+(q/kappa)^2
    Prod(S,T,Den);          -- Den = S*T
    Assign(J*N,S);
    if (NmJ mod 2)=0 then
      Prod(O.FourK0,S,Num); -- Num = 4*K0*nj  (or 2*K0*nj if TinyWFac)
    else
      Prod(O.FourK1,S,Num); -- Num = 4*K1*nj  (or 2*K1*nj if TinyWFac)
    end if;
    Quot(Num,Den,S);        -- S = Num/Den
  end Lqjn;

  procedure Lqjn(Q: in Flt; J,N: in Positive; S: in out FScalar) is
    --- not for loops
    O: QData;
  begin
    Init(Q,O);
    Lqjn(O,J,N,S);
  end Lqjn;

  procedure LqC2C(O: in QData; C1: in FPolyY; A1: in CAbs; C2: in out FPolyY) is
    --- assuming C2 is initialized to zero
    S,Tmp: FScalar;
  begin
    for N in 1 .. DegY loop
      if A1(N)/=Zero then
        for J in 1 .. DegY loop
          Lqjn(O,J,N,S);
          AddProd(S,C1(N),C2(J),Tmp);
        end loop;
      end if;
    end loop;
  end LqC2C;

  procedure LqC2E(O: in QData; A1: in CAbs; E2: in out ErrsY) is
    --- assuming E2 is initialized to zero
    RFourK1: constant Radius := MaxNorm(O.FourK1);
    A: Flt;
    S: FScalar;
  begin
    for N in 1 .. DegY loop
      if A1(N)/=Zero then
        A := A1(N)*NormFac(N);
        for J in 1 .. EDegY-1 loop
          Lqjn(O,J,N,S);
          E2(J) := E2(J)+A*MaxNorm(S);
        end loop;
        E2(EDegY) := E2(EDegY)+A*RFourK1*O.Sigma(EDegY-N);
      end if;
    end loop;
  end LqC2E;

  procedure LqE2E_Down(O: in QData; E1: in ErrsY; E2: in out ErrsY) is
    NegK0: constant Flt := -MaxNorm(O.K0);
    NegK1: constant Flt := -MaxNorm(O.K1);
    RNegQDivKap2: constant Flt := -Inf(O.QDivKap2);
    DEven: Boolean := True;
    Fac: Radius;
  begin
    for D in EFreqY loop -- frequency difference
      if DEven then
        Fac := NegK0/(RNegQDivKap2-Flt(D*D));
        DEven := False;
      else
        Fac := NegK1/(RNegQDivKap2-Flt(D*D));
        DEven := True;
      end if;
      for N in D+1 .. EDegY loop
        E2(N-D) := Fac*E1(N);
      end loop;
    end loop;
  end LqE2E_Down;

  procedure LqE2E_Up(O: in QData; E1: in ErrsY; E2: in out ErrsY) is
    Fac: constant Radius := MaxNorm(O.FourK1)*O.Sigma(1);
  begin
    for N in 2 .. EDegY loop
      if E1(N)/=Zero then
        E2(N) := E2(N)+Fac*E1(N-1);
      end if;
    end loop;
    E2(EDegY) := E2(EDegY)+Fac*E1(EDegY);
  end LqE2E_Up;

  procedure Init(P: in Cylinder; O: in out PData) is
    A: CAbs renames O.A;
    C: FPolyY renames P.C;
  begin
    O.ZC := Is_Zero(1,C);
    O.ZE := STrunc or else Is_Zero(P.E);
    if not O.ZC then
      A(0) := Zero;
      for I in 1 .. DegY loop A(I) := MaxNorm(C(I)); end loop;
    end if;
  end Init;

  procedure Lq(OQ: in QData; OP: in PData; P1: in Cylinder; P2: in out Cylinder) is
    --- check Q=0 and P1=0 first
    A1: CAbs renames OP.A;
    C1: FPolyY renames P1.C;
    C2: FPolyY renames P2.C;
    E1: ErrsY renames P1.E;
    E2: ErrsY renames P2.E;
  begin
    if P1.P=0 then raise Not_Implemented; end if;
    SetZero(1,P2);
    if not OP.ZC then
      LqC2C(OQ,C1,A1,C2);
      if Not_STrunc then LqC2E(OQ,A1,E2); end if;
    end if;
    if not OP.ZE then
      LqE2E_Down(OQ,E1,E2);
      LqE2E_Up(OQ,E1,E2);
    end if;
  end Lq;

  procedure Lq(Q: in Flt; P1: in Cylinder; P2: in out Cylinder) is
  begin
    if Q=Zero or else IsZero(P1) then
      Copy(P1,P2);
    else
      declare
        OQ: QData;
        OP: PData;
      begin
        Init(Q,OQ);
        Init(P1,OP);
        Lq(OQ,OP,P1,P2);
      end;
    end if;
  end Lq;

  function LMaxNorm(Q: Natural) return Radius is
    --- upper bound on ||L^q||
  begin
    if Q=0 then return One; end if;
    declare
      O: QData;
    begin
      Init(Flt(Q),O);
      return MaxNorm(O.K1)*(O.Sigma(0)+TwoOrFour*O.Sigma(1));
    end;
  end LMaxNorm;

  function LVarRho(P: Positive) return Radius is
    --- upper bound on ||L^q||^{1/q} for q>=p
  begin
    if STrunc then return One; end if;
    declare
      EPlus: constant Flt := Flt(356291)/Flt(131072); --- larger than e
      FltP:  constant Flt := Flt(P);
      R: Flt := Zero;
      S: Scalar;
      O: QData;
      use Flt_EF;
    begin
      Init(FltP,O);
      Quot(O.QDivKap2,Ell,S);    -- S=P/(Ell*Kappa^2), need S>e
      if Sup(S)<EPlus then raise Sorry; end if;
      R := O.Sigma(0)+TwoOrFour*O.Sigma(1);
      if R<One then R := One; end if;
      R := MaxNorm(O.K1)*R;
      return Exp(Log(R)/FltP);
    end;
  end LVarRho;

  procedure NegInvLap(A: in out Cylinder) is
    C: FPolyY renames A.C;
    M: Scalar;
  begin
    if A.P=0 then raise Not_Implemented; end if;
    if not IsZero(C(0)) then raise Inconsistent_Data; end if;
    for J in FreqY loop
      Prod(Flt(J*J),Kappa2,M);
      InvHelm(M,C(J));
    end loop;
    if not IsPoly(A) then
      declare
        K2: constant Radius := Inf(Kappa2);
        E: ErrsY renames A.E;
      begin
        E(1) := E(1)+E(0);
        E(0) := Zero;
        for J in 1 .. EDegY loop
          if E(J) /= Zero then
            E(J) := (E(J)/K2)/Flt(J*J);
          end if;
        end loop;
      end;
    end if;
  end NegInvLap;

  procedure NegInvLap(A1: in Cylinder; A2: in out Cylinder) is
  begin
    Copy(A1,A2);
    NegInvLap(A2);
  end NegInvLap;

  procedure Flip1(A: in out Cylinder) is
    C: FPolyY renames A.C;
  begin
    if A.P=0 then
      Flip(C(0));
    elsif not IsZero(C(0)) then
      raise Inconsistent_Data;
    end if;
    for J in FreqY loop Flip(C(J)); end loop;
  end Flip1;

  procedure Rotate(R: in Flt; A: in Cylinder; B: in out Cylinder) is
    AC: FPolyY renames A.C;
    BC: FPolyY renames B.C;
  begin
    if Ref/=0 then raise Undefined; end if;
    if A.P=0 then
      Rotate(R,AC(0),BC(0));
    elsif not IsZero(AC(0)) then
      raise Inconsistent_Data;
    end if;
    B.P := A.P;
    for J in FreqY loop Rotate(R,AC(J),BC(J)); end loop;
  end Rotate;

  procedure ProjEven(A: in out Cylinder) is
    C: FPolyY renames A.C;
  begin
    if A.P=0 then
      ProjEven(C(0));
    elsif not IsZero(C(0)) then
      raise Inconsistent_Data;
    end if;
    for J in FreqY loop ProjEven(C(J)); end loop;
  end ProjEven;

  procedure ProjOdd(A: in out Cylinder) is
    C: FPolyY renames A.C;
  begin
    if A.P=0 then
      ProjOdd(C(0));
    elsif not IsZero(C(0)) then
      raise Inconsistent_Data;
    end if;
    for J in FreqY loop ProjOdd(C(J)); end loop;
  end ProjOdd;

  procedure ProjASymm(N: in Positive; A: in out Cylinder) is
    C: FPolyY renames A.C;
  begin
    if A.P=0 then
      ProjASymm(N,C(0));
    elsif not IsZero(C(0)) then
      raise Inconsistent_Data;
    end if;
    for J in FreqY loop ProjASymm(N,C(J)); end loop;
  end ProjASymm;

  procedure ProjSymm(N: in Positive; A: in out Cylinder) is
    C: FPolyY renames A.C;
  begin
    if A.P=0 then
      ProjSymm(N,C(0));
    elsif not IsZero(C(0)) then
      raise Inconsistent_Data;
    end if;
    for J in FreqY loop ProjSymm(N,C(J)); end loop;
  end ProjSymm;

  procedure ProjConstant(A: in out Cylinder) is
    C: FPolyY renames A.C;
  begin
    if A.P=0 then
      ProjConstant(C(0));
    elsif not IsZero(C(0)) then
      raise Inconsistent_Data;
    end if;
    for J in FreqY loop ProjConstant(C(J)); end loop;
  end ProjConstant;

  procedure ValAtPi1(A: in Cylinder; V: in out Radial) is
    AC: FPolyY renames A.C;
    VC: RPoly renames V.C;
  begin
    if A.P=0 then
      V.P := 0;
      ValAtPi(AC(0),VC(0));
    else
      if not IsZero(AC(0)) then raise Inconsistent_Data; end if;
      V.P := 1;
      SetZero(VC(0));
    end if;
    for I in FreqY loop ValAtPi(AC(I),VC(I)); end loop;
    if Not_STrunc then
      declare
        AE: ErrsY renames A.E;
        VE: RErrs renames V.E;
      begin
        for I in reverse EFreqY loop VE(I) := AE(I); end loop;
      end;
    end if;
  end ValAtPi1;

  procedure Val1(A: in Cylinder; X: in Scalar; V: in out Radial) is
    AC: FPolyY renames A.C;
    VC: RPoly renames V.C;
  begin
    if A.P=0 then
      V.P := 0;
      Val(AC(0),X,VC(0));
    else
      if not IsZero(AC(0)) then raise Inconsistent_Data; end if;
      V.P := 1;
      SetZero(VC(0));
    end if;
    for I in FreqY loop Val(AC(I),X,VC(I)); end loop;
    if Not_STrunc then
      declare
        AE: ErrsY renames A.E;
        VE: RErrs renames V.E;
      begin
        for I in reverse EFreqY loop VE(I) := AE(I); end loop;
      end;
    end if;
  end Val1;

  procedure Plain_Val2(P: in Cylinder; Y: in Scalar; V: in out FScalar) is
    --- copy of CpCossins1.Val, but with Scalar argument
    C: FPolyY renames P.C;
    D: Integer := DegY;
  begin
    while IsZero(C(D)) loop --- QuasiDeg
      D := D-1;
      exit when D=0;
    end loop;
    if IsPoly(P) then
      if D=0 then
        if P.P=0 then Copy(C(0),V); else SetZero(V); end if;
        return;
      end if;
      SetZero(V);
    else
      if not IsReal(Y) then raise Domain_Error; end if;
      declare
        E: ErrsY renames P.E;
        ESum: Flt := Zero;
      begin
        for I in reverse EFreqY loop
          ESum := ESum+InvNormFac(I)*E(I);
        end loop;
        BallAt0(ESum,V);
      end;
    end if;
    declare
      CP,SP: RPoly;
      use Radials;
    begin
      CSPowers(Y,CP,SP);
      if P.P=0 then
        for I in reverse 1 .. D loop
          AddProd(CP(I),C(I),V);
        end loop;
        Add(C(0),V);
      else
        for I in reverse 1 .. D loop
          AddProd(SP(I),C(I),V);
        end loop;
      end if;
    end;
  end Plain_Val2;

  procedure Val2(A: in Cylinder; Y: in Scalar; V: in out FScalar) is
    KY: Scalar;
  begin
    Prod(Kappa,Y,KY);   -- Kappa*Y
    Plain_Val2(A,KY,V); -- does not know y-scaling
  end Val2;

  procedure Val(A: in Cylinder; X,Y: in Scalar; S: in out Scalar) is
    --- not suitable for loops
    KY: Scalar;
    V: Radial;
    use Radials;
  begin
    Val1(A,X,V);
    Prod(Kappa,Y,KY); -- Kappa*Y
    Val(V,KY,S);      -- does not know y-scaling
  end Val;

  procedure IVal(A: in Cylinder; XL,XR,YC: in Flt; S: in out Scalar; Steps: in Positive) is
    --- evaluate on [XL,XR] x {YC}
    Y: Scalar;
    V: FScalar;
  begin
    Assign(YC,Y);
    Val2(A,Y,V);
    IVal(V,XL,XR,S,Steps);
  end IVal;

  function IVal_Range(A: Cylinder; XL,XR,YC,ZMin,ZMax: Flt; Steps: Positive) return Boolean is
    --- check on [XL,XR] x {YC} whether all values are in (ZMin,ZMax)
    --- uses fewer Steps if possible
    OK: Boolean;
    Y: Scalar;
    V: FScalar;
  begin
    Assign(YC,Y);
    Val2(A,Y,V);
    OK := IVal_Range(V,XL,XR,ZMin,ZMax,Steps);
    if Verbosity>2 then
      Show1("YC =",YC,False);
      Show2(" XL,XR =",XL,XR);
    end if;
    return OK;
  end IVal_Range;

  procedure MultKappa(YL,YR: in Flt; KYL,KYR: out Flt) is
    S: Scalar;
  begin
    if YL>YR then raise Not_Implemented; end if;
    Prod(YL,Kappa,S);
    KYL := Inf(S);
    Prod(YR,Kappa,S);
    KYR := Sup(S);
  end MultKappa;

  procedure JValAt0(A: in Cylinder; YL,YR: in Flt; MinInf,MaxInf,MinSup,MaxSup: out Flt; Steps: in Positive) is
    --- evaluate on {0} x [YL,YR]
  begin
    JVal(A,Zero,YL,YR,MinInf,MaxInf,MinSup,MaxSup,Steps);
  end JValAt0;

  procedure JValAtPi(A: in Cylinder; YL,YR: in Flt; MinInf,MaxInf,MinSup,MaxSup: out Flt; Steps: in Positive) is
     --- evaluate on {Pi} x [YL,YR]
    KYL,KYR: Flt;
    V: Radial;
    use Radials;
  begin
    ValAtPi1(A,V);
    MultKappa(YL,YR,KYL,KYR);
    IVal(V,KYL,KYR,MinInf,MaxInf,MinSup,MaxSup,Steps); -- does not know y-scaling
  end JValAtPi;

  procedure JVal(A: in Cylinder; XC,YL,YR: in Flt; MinInf,MaxInf,MinSup,MaxSup: out Flt; Steps: in Positive) is
    --- evaluate on {XC} x [YL,YR]
    KYL,KYR: Flt;
    X: Scalar;
    V: Radial;
    use Radials;
  begin
    Assign(XC,X);
    Val1(A,X,V);
    MultKappa(YL,YR,KYL,KYR);
    IVal(V,KYL,KYR, MinInf,MaxInf,MinSup,MaxSup,Steps); -- does not know y-scaling
  end JVal;

  function JVal_Range(A: Cylinder; XC,YL,YR,ZMin,ZMax: Flt; Steps: Positive) return Boolean is
    --- check on {XC} x [YL,YR] whether all values are in (ZMin,ZMax)
    OK: Boolean;
    KYL,KYR: Flt;
    X: Scalar;
    V: Radial;
    use Radials;
  begin
    Assign(XC,X);
    Val1(A,X,V);
    MultKappa(YL,YR,KYL,KYR);
    OK := IVal_Range(V,KYL,KYR,ZMin,ZMax,Steps);
    if Verbosity>2 then
      Show1("XC =",XC,False);
      Show2(" YL,YR =",YL,YR);
    end if;
    return OK;
  end JVal_Range;

  function RVal_Range(A: Cylinder; XL,XR,YL,YR,ZMin,ZMax: Flt; IL,JL: Positive) return Boolean is
    --- evaluate along boundary of [XL,XR] x [YL,YR]
    --- IL and JL are guessed number of X-steps and Y-steps
    OK1,OK2,OK3,OK4: Boolean := True;
    PC: Protected_Counter(1);
    procedure BTRL is
      J: Positive;
    begin
      loop
        PC.Next(J);
        case J is
          when 1 => OK1 := IVal_Range(A,XL,XR,YL,ZMin,ZMax,IL); --- bottom: [XL,XR] x {YL}
          when 2 => OK2 := IVal_Range(A,XL,XR,YR,ZMin,ZMax,IL); --- top:    [XL,XR] x {YR}
          when 3 => OK3 := JVal_Range(A,XR,YL,YR,ZMin,ZMax,JL); --- right:  {XR} x [YL,YR]
          when 4 => OK4 := JVal_Range(A,XL,YL,YR,ZMin,ZMax,JL); --- left:   {XL} x [YL,YR]
          when others => exit;
        end case;
        exit when not (OK1 and then OK2 and then OK3 and then OK4);
      end loop;
    end BTRL;
    NT: constant Natural := Reserve_Tasks(3,Allow_Parallel);
  begin
    if Verbosity>1 then
      Show2("  XL   XR:",XL,XR);
      Show2("  YL   YR:",YL,YR);
      Show2("ZMin ZMax:",ZMin,ZMax);
    end if;
    if NT=0 then
      BTRL;
    else
      declare
        Task_Error: Boolean := False;
        task type BTRL_Task_Type is end BTRL_Task_Type;
        task body BTRL_Task_Type is
        begin
          Proper_Rounding;
          BTRL;
        exception
          when others => Task_Error := True; raise;
        end BTRL_Task_Type;
        ----
        BTRL_Task: array(1 .. NT) of BTRL_Task_Type;
        pragma Warnings (Off,BTRL_Task);
      begin
        BTRL;
        if Task_Error then raise Task_Exception; end if;
      end;
      Free_Tasks(NT);
    end if;
    return (OK1 and then OK2 and then OK3 and then OK4);
  end RVal_Range;

  procedure CheckMin(A: in Cylinder; XC,XL,XR,YC,YL,YR: in Flt; IL,JL: in Positive) is
    ZMax: constant Flt := 1.0E+999;
    OK: Boolean;
    ZMin: Flt;
    X,Y,Z: Scalar;
  begin
    if XC>XR or XC<XL or YC>YR or YC<YL then raise Sorry; end if;
    Assign(XC,X);
    Assign(YC,Y);
    Val(A,X,Y,Z);
    ZMin := Sup(Z);
    OK := RVal_Range(A,XL,XR,YL,YR,ZMin,ZMax,IL,JL);
    if OK then
      Show1("OK: Min check succeeded. MaxC =", ZMin);
    else
      Show1("Error: Min check failed. MaxC =", ZMin);
    end if;
    if Verbosity>2 then
      Show2("  XL,   XR:",XL,XR);
      Show2("  YL,   YR:",YL,YR);
    end if;
  end CheckMin;

  procedure CheckMax(A: in Cylinder; XC,XL,XR,YC,YL,YR: in Flt; IL,JL: in Positive) is
    ZMin: constant Flt := -1.0E+999;
    OK: Boolean;
    ZMax: Flt;
    X,Y,Z: Scalar;
  begin
    if XC>XR or XC<XL or YC>YR or YC<YL then raise Sorry; end if;
    Assign(XC,X);
    Assign(YC,Y);
    Val(A,X,Y,Z);
    ZMax := Inf(Z);
    OK := RVal_Range(A,XL,XR,YL,YR,ZMin,ZMax,IL,JL);
    if OK then
      Show1("OK: Max check succeeded. MinC =", ZMax);
    else
      Show1("Error: Max check failed. MinC =", ZMax);
    end if;
    if Verbosity>2 then
      Show2("  XL,   XR:",XL,XR);
      Show2("  YL,   YR:",YL,YR);
    end if;
  end CheckMax;

  procedure LevelStat(A: in Cylinder; Y,Cut: in Flt; Steps: in Positive) is
    --- determine values above/below Cut for fixed Y
    S: Scalar;
    P: FScalar;
  begin
    Assign(Y,S);
    Val2(A,S,P);
    LevelStat(P,Cut,Steps);
  end LevelStat;

  procedure Near_Cassini_Injective(R: in Flt) is
    --- checks whether Near_Cassini is one-to-one
    NLast: constant Positive := (LDeg+1)/2;
    M: constant Positive := NLast+1;
    RM: constant Radius := NPower(R,M);    --- R^M
    RQ: constant Radius := NegOne/(R-One); --- 1/(1-R)
    A: Flt;
    use Flt_EF;
  begin
    A := (One+R)*Two*RM*RQ;
    if A<=One then
      Show1("psi check OK. A =",A);
    else
      Show1("Error: psi check Failed. A =",A);
    end if;
    A := (One+R*R)*RQ*(Sqrt(RQ)*FC(M,R)+RM*Sqr(RQ));
    if A<=One then
      Show1("chi check OK:",A);
    else
      Show1("Error: chi check Failed. A =",A);
    end if;
  end Near_Cassini_Injective;

  ----------------------------------------- LTaylors

  procedure Mult(S: in Scalar; P: in out LTaylor) is
    C: Poly1 renames P.C;
  begin
    for I in Power loop Mult(S,C(I)); end loop;
    if Not_STrunc then P.E := MaxNorm(S)*P.E; end if;
  end Mult;

  procedure Prod(S: in Scalar; P1: in LTaylor; P2: in out LTaylor) is
    C1: Poly1 renames P1.C;
    C2: Poly1 renames P2.C;
  begin
    for I in Power loop Prod(S,C1(I),C2(I)); end loop;
    if Not_STrunc then P2.E := MaxNorm(S)*P1.E; end if;
  end Prod;

  procedure Apply(T: in LTaylor; A: in Cylinder; B: in out Cylinder) is
    TC: Poly1 renames T.C;
    R: Flt;
    OQ: QData;
    OP: PData;
    LqA: Cylinder;
  begin
    if IsZero(A) then Copy(A,B); return; end if;
    Init(A,OP);
    Prod(TC(0),A,B);
    for Q in 1 .. LDeg loop
      if not IsZero(TC(Q)) then
        R := Flt(Q);
        Init(R,OQ);
        Lq(OQ,OP,A,LqA);  -- L^q(A)
        Mult(TC(Q),LqA);  -- TC(Q)*L^q(A)
        Add(LqA,B);       -- B := B+TC(Q)*L^q(A)
      end if;
    end loop;
    if not IsPoly(T) then
      R := T.E*MaxNorm(A);
      AddBall(R,B);
    end if;
  end Apply;

  -----------------------------

  function FA(M: Natural; X: Radius) return Flt is
    --- upper bound on Sum_{n>=m} x^n
    Y: constant Radius := NegOne/(X-One);
  begin
    if M=0 then return Y; end if;  -- 1/(1-X)
    return NPower(X,M)*Y;          -- (X**M)/(1-X)
  end FA;

  function FB(M: Natural; X: Radius) return Flt is
    --- upper bound on Sum_{n>=m} B_n x^n
    Y: constant Radius := NegOne/(X-One);
    use Flt_EF;
  begin
    if M=0 then return Sqrt(Y); end if;  -- 1/Sqrt(1-X)
    return NPower(X,M)*Sqrt(Y);          -- (X**M)/Sqrt(1-X)
  end FB;

  function FC(M: Natural; X: Radius) return Flt is
    --- upper bound on Sum_{n>=m} C_n x^n
    C: constant Flt := MaxNorm(CVal(M));  --- requires M <= LDeg
    Y: constant Radius := NegOne/(X-One);
    use Flt_EF;
  begin
    if M=0 then return Y*Sqrt(Y); end if;        -- 1/(1-X)^{3/2}
    return C*NPower(X,M)*Y*(One+X*Y/Flt(2*M+1)); -- see the notes
  end FC;

  procedure PowerNegOne1(R: in Scalar; P: in out LTaylor; NLast: in Positive; PolyPsi: in Boolean) is
    --- |f|^2 for f=1/(1-r\zeta)
    RLast: constant Positive := 2*NLast;   --- powers of R needed
    S: Scalar;
    RP: array(0 .. RLast) of Scalar;
    F: FScalar;
  begin
    Copy(SOne,RP(0));
    Copy(R,RP(1));
    for N in 2 .. RLast loop Prod(RP(N/2),RP(N-N/2),RP(N)); end loop; -- powers of R

    SetZero(P);
    for N in 0 .. NLast loop
      for M in 0 .. N loop
        if M=N then
          Copy(RP(N+M),S);
        else
          Prod(STwo,RP(N+M),S);
        end if;
        SetZero(F);
        AddCoeff(N-M,S,F);            -- F = S*cos((N-M).)
        AddCoeff(N+M,F,P);            -- add F*L^{N+M} to P
      end loop;
    end loop;

    if not (STrunc or PolyPsi) then
      if NLast<LDeg then raise Not_Implemented; end if; -- P.E inludes L^q with q>LDeg
      declare
        M: constant Positive := NLast+1;
        A:   constant Radius := MaxNorm(R);
        Rho: constant Radius := RMax(RadX,LVarRho(M));
        B:   constant Radius := A*(Rho**2);
      begin
        if B>=One then raise Domain_Error; end if;
        P.E := FA(M,B)*FA(0,A)+FA(0,B)*FA(M,A);
      end;
    end if;
  end PowerNegOne1;

  procedure PowerNegOne2(R: in Scalar; P: in out LTaylor; NLast: in Positive; PolyPsi: in Boolean) is
    --- |f|^2 for f=1/(1-r\zeta^2)
    RLast: constant Positive := 2*NLast;   --- powers of R needed
    S: Scalar;
    RP: array(0 .. RLast) of Scalar;
    F: FScalar;
  begin
    Copy(SOne,RP(0));
    Copy(R,RP(1));
    for N in 2 .. RLast loop Prod(RP(N/2),RP(N-N/2),RP(N)); end loop; -- powers of R

    SetZero(P);
    for N in 0 .. NLast loop
      for M in 0 .. N loop
        if M=N then
          Copy(RP(N+M),S);
        else
          Prod(STwo,RP(N+M),S);
        end if;
        SetZero(F);
        AddCoeff(2*(N-M),S,F);            -- F = S*cos(2(N-M).)
        AddCoeff(2*(N+M),F,P);            -- add F*L^{2(N+M)} to P
      end loop;
    end loop;

    if not (STrunc or PolyPsi) then
      if 2*NLast<LDeg then raise Not_Implemented; end if;
      declare
        M: constant Positive := NLast+1;
        A:   constant Radius := MaxNorm(R);
        Rho: constant Radius := RMax(RadX,LVarRho(2*M));
        B:   constant Radius := A*(Rho**4);
      begin
        if B>=One then raise Domain_Error; end if;
        P.E := FA(M,B)*FA(0,A)+FA(0,B)*FA(M,A);
      end;
    end if;
  end PowerNegOne2;

  procedure PowerNegThreeHalf(R: in Scalar; P: in out LTaylor; NLast: in Positive; PolyPsi: in Boolean) is
    --- |f|^2 for f=(1-r\zeta^2)^{-3/2}
    --- first initialize global CVal
    RLast: constant Positive := 2*NLast;   --- powers of R needed
    S: Scalar;
    RP: array(0 .. RLast) of Scalar;
    F: FScalar;
  begin
    Copy(SOne,RP(0));
    Copy(R,RP(1));
    for N in 2 .. RLast loop Prod(RP(N/2),RP(N-N/2),RP(N)); end loop; -- powers of R

    SetZero(P);
    if LMax<NLast then raise Sorry; end if; -- increase LMax
    for N in 0 .. NLast loop
      for M in 0 .. N loop
        Prod(CVal(N),CVal(M),S);            -- S := |{-3/2 choose N}|*|{-3/2 choose M}|
        Mult(RP(N+M),S);                    -- S := S * R^{N+M}
        if M/=N then Mult(Two,S); end if;
        SetZero(F);
        AddCoeff(2*(N-M),S,F);              -- F = S*cos(2(N-M).)
        AddCoeff(2*(N+M),F,P);              -- add F*L^{2(N+M)} to P
      end loop;
    end loop;

    if not (STrunc or PolyPsi) then
      if 2*NLast<LDeg then raise Not_Implemented; end if;
      declare
        M: constant Positive := NLast+1;
        A:   constant Radius := MaxNorm(R);
        Rho: constant Radius := RMax(RadX,LVarRho(2*M));
        B:   constant Radius := A*(Rho**4);
      begin
        if B>=One then raise Domain_Error; end if;
        P.E := FC(M,B)*FC(0,A)+FC(0,B)*FC(M,A);
      end;
    end if;
  end PowerNegThreeHalf;

  ------------------- examples

  procedure Disk_Iso(R: in Flt; T: in out LTaylor; PolyPsi: in Boolean := False) is
    --- |psi'|^2 for psi'=1/(1+r\zeta)^2
    --- Example 1
    NLast: constant Positive := LDeg; --- expansion cutoff
    S: Scalar;
    Tmp: LTaylor;
  begin
    Assign(-R,S);
    PowerNegOne1(S,Tmp,NLast,PolyPsi); -- |f|^2 for f=1/(1+r\zeta)
    Prod(Tmp,Tmp,T);                   -- |f|^4
  end Disk_Iso;

  procedure Disk_Iso_Plus(R: in Flt; T: in out LTaylor; PolyPsi: in Boolean := False) is
  begin
    Disk_Iso(R,T,PolyPsi);
    LMult(2,T);
  end Disk_Iso_Plus;

  procedure InvEllipse(A: in Flt; T: in out LTaylor; PolyPsi: in Boolean := False) is
    --- |psi'|^2 for psi'=(1+r\zeta^2)/(1-r\zeta^2)^2
    --- Example 6
    NLast: constant Positive := (LDeg+1)/2; --- expansion cutoff (notice: 2 orders per step)
    R,S: Scalar;
    Tmp: LTaylor;
    C: Poly1 renames Tmp.C;
  begin
    Assign(A,R);
    PowerNegOne2(R,Tmp,NLast,PolyPsi); -- |f|^2 for f=1/(1-r\zeta^2)
    Prod(Tmp,Tmp,T);                   -- |f|^4

    SetZero(Tmp);
    Monom(0,SOne,C(0)); -- L^0 coeff: 1
    Prod(Two,R,S);
    Monom(2,S,C(2));    -- L^2 coeff: 2*Rcos(2.)
    Prod(R,R,S);
    Monom(0,S,C(4));    -- L^4 coeff: R^2
    Mult(Tmp,T);        -- multiply by |1+r\zeta^2|^2 = 1+2*Rcos(2.)L^2+(R^2)L^4
  end InvEllipse;

  procedure InvEllipse_Plus(A: in Flt; T: in out LTaylor; PolyPsi: in Boolean := False) is
  begin
    InvEllipse(A,T,PolyPsi);
    LMult(2,T);
  end InvEllipse_Plus;

  procedure Cassini(A: in Flt; T: in out LTaylor) is
    NLast: constant Positive := (LDeg+1)/2; --- expansion cutoff (notice: 2 orders per step)
    R: Scalar;
  begin
    Assign(A,R);
    PowerNegThreeHalf(R,T,NLast,False); -- |\psi'|^2 for \psi'=(1-r\zeta^2)^{-3/2}
  end Cassini;

  procedure Cassini_Plus(A: in Flt; T: in out LTaylor) is
  begin
    Cassini(A,T);
    LMult(2,T);
  end Cassini_Plus;

  procedure Near_Cassini(A: in Flt; T: in out LTaylor; NLast: in Positive) is
    R: Scalar;
  begin
    Assign(A,R);
    PowerNegThreeHalf(R,T,NLast,True); -- |\psi'|^2 for \psi'=(1-r\zeta^2)^{-3/2}
  end Near_Cassini;

  procedure Near_Cassini_Plus(A: in Flt; T: in out LTaylor; NLast: in Positive) is
  begin
    Near_Cassini(A,T,NLast);
    LMult(2,T);
  end Near_Cassini_Plus;

  ----------------------------------------- modes

  procedure Norm(M: in CylinderMode; S: in out Scalar) is
  begin
    if M.ErrX or else M.ErrY then
      Copy(SOne,S);  --- error modes have norm 1
    else
      declare
        W,Tmp: Scalar;
      begin
        Assign(M.R,W);
        IPower(abs(M.K),W,S,Tmp);
        Mult(NormFac(M.J),S,Tmp);
      end;
    end if;
  end Norm;

  procedure MaxNorm(M: in CylinderMode; S: out Radius) is
  begin
    if M.ErrX or else M.ErrY then
      S := One;      --- error modes have norm 1
    else
      S := NPower(M.R,abs(M.K))*NormFac(M.J);
    end if;
  end MaxNorm;

  procedure Assign(M: in CylinderMode; Q: in out Cylinder) is
    --- assuming that M is fine
    R: Radius renames M.R;
    C: FPolyY renames Q.C;
  begin
    SetZero(1,Q);
    for I in FreqY loop C(I).R := R; end loop;
    if M.ErrX then
      C(M.J).E(M.K) := InvNormFac(M.J);
    elsif M.ErrY then
      Q.E(M.J) := One;
    else
      Copy(SOne,C(M.J).C(M.K));
    end if;
  end Assign;

  procedure AddProd(M: in CylinderMode; S: in Scalar; Q: in out Cylinder; Check: in Boolean) is
    pragma Unreferenced(Check);
  begin
    if not IsZero(S) then
      if M.ErrX or else M.ErrY then raise Not_Implemented; end if;
      AddCoeff(M.K,S,Q.C(M.J));
    end if;
  end AddProd;

  procedure Extract_Coeff(M: in CylinderMode; Q: in out Cylinder; S: in out Scalar) is
    --- assuming M.Err=False
    --- S encloses the coefficient
  begin
    if not M.Simple then raise Not_Implemented; end if;
    SetZero(S);
    Swap(Q.C(M.J).C(M.K),S);
  end Extract_Coeff;

  procedure Fail(M: in CylinderMode; Q: in out Cylinder; S: in out Scalar) is
    pragma Unreferenced(M);
    pragma Unreferenced(Q);
    pragma Unreferenced(S);
  begin
    raise Not_Implemented;
  end Fail;

  procedure Extract(M: in CylinderMode; Q: in out Cylinder; S: in out Scalar; Check: in Boolean) is
    --- assuming M is consistent
    pragma Unreferenced(Check);
  begin
    if M.ErrX or else M.ErrY then raise Not_Implemented; end if;
    Extract_Coeff(M,Q,S);
  end Extract;

  procedure Show(M: in CylinderMode; NewLine: in Boolean := True) is
    pragma Unreferenced(NewLine);
  begin
    Show1(" ErrX = ",M.ErrX);
    Show1(" ErrY = ",M.ErrY);
    Show1("    J = ",M.J);
    Show1("    K = ",M.K);
  end Show;

  function SameType(M1,M2: CylinderMode) return Boolean is
  begin
    return (M1.Simple=M2.Simple);
  end SameType;

  function Symm_Check(M: Cylindermode) return Boolean is
    K: EFreqX renames M.K;
  begin
    if Full then return True; end if;                      -- all modes
    --- reflection
    if Ref= 1 and then K <0 then return False; end if;     -- Even
    if Ref=-1 and then K>=0 then return False; end if;     -- Odd
    --- rotation
    if Symm=0 then return K=0; end if;                     -- radial
    if Symm=1 then return True; end if;                    -- no symmetry
    if M.ErrY or else M.ErrX then return True; end if;     -- Errors ok
    if Symm>1 then return (K mod AbsRot)=0; end if;        -- Symm>1 rotation symm
    return (K mod (2*AbsRot))=AbsRot;                      -- Symm<0 rotation antisymm
  end Symm_Check;

  function Make(RP: Radius; Siz: Positive; Simple,Trunc: Boolean) return CylinderModes is
    Exclude_Errs: constant Boolean := Trunc or STrunc;
    Include_Errs: constant Boolean := not Exclude_Errs;
    S: constant Boolean := Simple or Exclude_Errs;
    Size2: constant Positive  := Siz/2;
    SizeY: constant Positive  := 2*Size2;   --- even
    SizeX: Natural;

    procedure Build(M: out CylinderModes; Count: out Natural) is

      procedure Set_Mode(J: in Positive; K: in Integer; ErrY,ErrX: in Boolean) is
        Tmp: constant CylinderMode := (RP,J,K,ErrY,ErrX,S);
      begin
         if Symm_Check(Tmp) then
            Count := Count+1;
            if Count <= M'Last then M(Count) := Tmp; end if;
         end if;
      end Set_Mode;

      function SizX(J: Positive) return Natural is
      begin
         if Symm=0 then
            return 0;
         else
            return IMin(IMax(((SizeY-J)/Symm)*Symm,Size2),DegX);
         end if;
      end SizX;

    begin
      Count := 0;
      for J in 1 .. SizeY loop
         SizeX := SizX(J);
         for K in -SizeX .. SizeX loop
            Set_Mode(J,K,False,False); --- coefficient
         end loop;
      end loop;
      if Include_Errs then
         Set_Mode(SizeY+1,0,True,False);
         for J in 1 .. SizeY loop
            SizeX := SizX(J);
            Set_Mode(J,SizeX+1,False,True);
            Set_Mode(J,-SizeX-1,False,True);
         end loop;
      end if;
    end Build;

    Count: Natural;
  begin
    declare
      M: CylinderModes(1..1);
    begin
      Build(M,Count);
    end;
    declare
      M: CylinderModes(1..Count);
    begin
      Build(M,Count);
      return M;
    end;
  end Make;

  ------------------------------------ plotting
  ------------------------------------ plotting

  procedure Plot(P: in Point; F: in File_Type) is
    package Txt_IO renames Ada.Text_IO;
    Digs: constant Positive := 5;
  begin
    Flt_IO.Put(F,P.X,3,Digs,3);
    Txt_IO.Put(F," ");
    Flt_IO.Put(F,P.Y,3,Digs,3);
    Txt_IO.Put(F," ");
    Flt_IO.Put(F,P.Z,3,Digs,3);
    New_Line(F);
  end Plot;

  function ApproxVal(X,Y: Flt; A: Cylinder) return Flt is
    KY: constant Flt := RKappa*Y;
    C: FPolyY renames A.C;
    Z: Flt := Zero;
    use Flt_EF;
  begin
    for J in reverse 1 .. DegY loop
      Z := Z+ApproxVal(X,C(J))*Sin(Flt(J)*KY);
    end loop;
    return Z;
  end ApproxVal;

  function ApproxGrad2(X,Y: Flt; A: Cylinder) return Flt is
    --- |gradient A|^2(x,y)
    C: FPolyY renames A.C;
    KJ,KJY: Flt;
    Z: Flt := Zero;
    use Flt_EF;
  begin
    for J in reverse 1 .. DegY loop
      KJ := RKappa*Flt(J);
      KJY := KJ*Y;
      Z := Z+Sqr(ApproxValDer(X,C(J))*Sin(KJY))+Sqr(ApproxVal(X,C(J))*KJ*Cos(KJY));
    end loop;
    return Z;
  end ApproxGrad2;

  function WPhi1(Y: Flt) return Flt is
    -- Phi', plain annulus
    use Flt_EF;
  begin
    return Exp(-Two*Y);
  end WPhi1;

  function WPhi2(X,Y,R: Flt; N: Integer) return Flt is
    --- Phi', Example 3
    NX: constant Flt := Flt(N)*X;
    NY: constant Flt := Flt(N)*Y;
    WPsi: Flt;
    use Flt_EF;
  begin
    WPsi := One-Two*R*Cos(NX)*Exp(-NY)+R*R*Exp(-Two*NY);
    return Exp(-Two*Y)*WPsi;
  end WPhi2;

  function WPhi3(X,Y,R: Flt) return Flt is
    --- Phi', Example 1
    WPsi: Flt;
    use Flt_EF;
  begin
    WPsi := One+Two*R*Cos(X)*Exp(-Y)+R*R*Exp(-Two*Y);
    WPsi := One/Sqr(WPsi);
    return Exp(-Two*Y)*WPsi;
  end WPhi3;

  function WPhi4(X,Y,R: Flt) return Flt is
    --- Phi', Example 7
    WPsi: Flt;
    use Flt_EF;
  begin
    WPsi := One-Two*R*Cos(Two*X)*Exp(-Two*Y)+R*R*Exp(-Four*Y);
    WPsi := Exp(-ThreeQuarters*Log(WPsi));
    return Exp(-Two*Y)*WPsi;
  end WPhi4;

  function WPhi5(X,Y,R: Flt) return Flt is
    --- Phi', Example 6
    Num,Den,WPsi: Flt;
    use Flt_EF;
  begin
    Num := One+Two*R*Cos(Two*X)*Exp(-Two*Y)+R*R*Exp(-Four*Y);
    Den := One-Two*R*Cos(Two*X)*Exp(-Two*Y)+R*R*Exp(-Four*Y);
    WPsi := Num/(Den*Den);
    return Exp(-Two*Y)*WPsi;
  end WPhi5;

  ------------------------------

  function Phi1(P: Point) return Point is
    --- e^{iz}
    use Flts.Complexs;
    Z: constant Complex := (P.X,P.Y);
    Zeta: Complex;
  begin
    Zeta := Exp(CImag*Z);
    return (Zeta.Re,Zeta.Im,P.Z);
  end Phi1;

  function Phi2(P: Point; R: Flt; N: Integer) return Point is
    --- psi=Zeta-(r/(n+1))Zeta^{n+1},  Zeta=e^{iz}
    --- Example 3
    use Flts.Complexs;
    Z: constant Complex := (P.X,P.Y);
    Zeta,Psi: Complex;
  begin
    Zeta := Exp(CImag*Z);
    Psi := Zeta-(R/Flt(N+1))*Zeta**(N+1);
    return (Psi.Re,Psi.Im,P.Z);
  end Phi2;

  function Phi3(P: Point; R: Flt) return Point is
    --- psi=C*(Zeta+r)/(1+r Zeta),  Zeta=e^{iz}
    --- Example 1
    use Flts.Complexs;
    C: constant Flt := One/Sqr(One-R*R);
    Z: constant Complex := (P.X,P.Y);
    Zeta,Psi: Complex;
  begin
    Zeta := Exp(CImag*Z);
    Psi := C*((Zeta+R*COne)/(COne+R*Zeta));
    return (Psi.Re,Psi.Im,P.Z);
  end Phi3;

  function Phi4(P: Point; R: Flt) return Point is
    --- psi=Zeta/sqrt(1-r Zeta^2), Zeta=e^{iz}
    --- Example 7, Cassini
    use Flt_EF,Flts.Complexs;
    Z: constant Complex := (P.X,P.Y);
    Zeta,Psi: Complex;
  begin
    Zeta := Exp(CImag*Z);
    Psi := Zeta/Sqrt(COne-R*(Zeta*Zeta));
    return (Psi.Re,Psi.Im,P.Z);
  end Phi4;

  function Phi5(P: Point; R: Flt) return Point is
    --- psi=Zeta/(1-r Zeta^2), Zeta=e^{iz}
    --- Example 6, inverted ellipse
    use Flt_EF,Flts.Complexs;
    Z: constant Complex := (P.X,P.Y);
    Zeta,Psi: Complex;
  begin
    Zeta := Exp(CImag*Z);
    Psi := Zeta/(COne-R*(Zeta*Zeta));
    return (Psi.Re,Psi.Im,P.Z);
  end Phi5;

  function Phi6(P: Point; R: Flt) return Point is
    --- Example 7', Near_Cassini
    use Flt_EF,Flts.Complexs;
    NLast: constant Positive := (LDeg+1)/2; --- expansion cutoff
    Z: constant Complex := (P.X,P.Y);
    Zeta: constant Complex := Exp(CImag*Z);
    T: constant Complex := R*(Zeta*Zeta);   --- main expansion in powers of T
    C: Flt;
    Prod,Psi: Complex := COne;
  begin
    for N in 1 .. NLast loop
      C := Flt(2*N-1)/Flt(2*N);  -- new combinatorial factor
      Prod := (C*Prod)*T;        -- (-R*Zeta^2)^N {-1/2 choose N}
      Psi := Psi+Prod;
    end loop;
    Psi := Zeta*Psi;
    return (Psi.Re,Psi.Im,P.Z);
  end Phi6;

end Cylinders;
