with Ada.Integer_Text_IO, Globals, Strings, N_M, Protected_Counters;
use Globals, Strings, Protected_Counters;

pragma Elaborate_All (Ada.Integer_Text_IO,Globals,Strings,N_M,Protected_Counters);

package body Flts.Minima is

  ----------------------- needed by N_M

  procedure Proper_Rounding is
  begin
    null;
  end  Proper_Rounding;

  procedure Center(S: in out Flt) is
    pragma Unreferenced (S);
  begin
    null;
  end Center;

  procedure Copy(S1: in Flt; S2: in out Flt) is
  begin
    S2 := S1;
  end Copy;

  procedure Swap(S1,S2: in out Flt) is
    R1: constant Flt := S1;
  begin
    S1 := S2;
    S2 := R1;
  end Swap;

  procedure Add(S1: in Flt; S2: in out Flt) is
  begin
    S2 := S2+S1;
  end Add;

  procedure Sub(S1: in Flt; S2: in out Flt) is
  begin
    S2 := S2-S1;
  end Sub;

  procedure Sum(S1,S2: in Flt; S3: in out Flt) is
  begin
    S3 := S1+S2;
  end Sum;

  procedure Diff(S1,S2: in Flt; S3: in out Flt) is
  begin
    S3 := S1-S2;
  end Diff;

  procedure Mult(R: in Flt; S: in out Flt) is
  begin
    S := S*R;
  end Mult;

  package Nelder_Mead is new N_M (Dim => 2, Scalar => Flt);
  use Nelder_Mead;

  ---------------------------- Local minima

  procedure Fill(Z: out Values) is
    Task_Error: Boolean := False;
    PC: Protected_Counter(0);

    procedure Fill_Loop is
      N: Natural;
      X,Y: Flt;
    begin
      loop
        PC.Next(N);
        exit when N>YL;
        Y := Yof(N);
        for M in MRange loop
          X := Xof(M);
          Z(M,N) := Fun(X,Y);
        end loop;
      end loop;
    exception
      when others => Task_Error := True; raise;
    end Fill_Loop;

    NT: constant Natural := Reserve_Tasks(YL,Fill_Parallel);
  begin
    if NT=0 then
      Fill_Loop;
    else
      declare
        task type Fill_Task_Type is end Fill_Task_Type;
        task body Fill_Task_Type is begin Fill_Loop;  end Fill_Task_Type;
        ---
        Fill_Task: array(1 .. NT) of Fill_Task_Type;
        pragma Warnings (Off,Fill_Task);
      begin
        Fill_Loop;
        if Task_Error then raise Task_Exception; end if;
      end;
      Free_Tasks(NT);
    end if;
  end Fill;

  function Fill return Values is
    Z: Values;
  begin
    Fill(Z);
    return Z;
  end Fill;

  function Rough_Local_Minima(Z: Values; Dist: Positive) return Flt_Vec is
    --- stays away from the boundary
    function IsMin(M0,N0: Integer) return Boolean is
      MF: constant Integer := M0-Dist;
      ML: constant Integer := M0+Dist;
      NF: constant Integer := N0-Dist;
      NL: constant Integer := N0+Dist;
      Z0: Flt renames Z(M0,N0);
    begin
      for M in MF .. ML loop
        for N in NF .. NL loop
          if Z(M,N)<Z0 then return False; end if;
        end loop;
      end loop;
      return True;
    end IsMin;
    ML: constant Integer := XL-Dist;
    NL: constant Integer := YL-Dist;
    Last: Integer := 0;
    X,Y,V: Flt;
    Found: Flt_Vec(1 .. 3*XL);
  begin
    for M in Dist .. ML loop
      for N in Dist .. NL loop
        if IsMin(M,N) then
          V := Z(M,N);
          if abs(V) >= Cutoff then
            X := Xof(M);
            Y := Yof(N);
            Found(Last+1) := X;
            Found(Last+2) := Y;
            Found(Last+3) := V;
            Last := Last+3;
          end if;
        end if;
      end loop;
    end loop;
    if Last <= 3 then
      Show0(Strng(Last/3,3) & " rough extremum found");
    else
      Show0(Strng(Last/3,3) & " rough extrema found");
    end if;
    return Found(1 .. Last);
  end Rough_Local_Minima;

  function Rough_Local_Minima(Dist: Positive) return Flt_Vec is
    --- stays away from the boundary
    Z: Values;
  begin
    Fill(Z);
    return Rough_Local_Minima(Z,Dist);
  end Rough_Local_Minima;

  procedure Improve_Min(P: in out VecFltVec; Eps: in Radius) is
    MaxCount: constant Positive := 64;
    Count: Integer := 0;
    YMin: Flt := Empty;
    ---
    function F(X: Vec) return Flt is begin return Fun(X(1),X(2)); end F;
    ---
    function Done(Q: VecFltVec) return Boolean is
      Y1: Flt renames Q(1).Y;
      Fertig: Boolean;
      Err: Radius;
    begin
      if Count=0 then
        Fertig := False;
      else
        Err := abs(YMin-Y1);
        Fertig := Err<Eps;
      end if;
      YMin := Y1;
      Count := Count+1;
      if Count=MaxCount then
        Show1("giving up, err =",Err);
      elsif Fertig and then Verbosity>2 then
        Show1("done, err =",Err);
      end if;
      return Fertig;
    end Done;
    ---
    procedure Find_Min is new FindMin (F => F, Done => Done);
  begin
    Find_Min(P);
  end Improve_Min;

  procedure Improve_Min(V: in out Flt_Vec; Eps: in Radius) is
    HalfDX: constant Flt := Half*(XLast-XFirst)/Flt(XL);
    HalfDY: constant Flt := Half*(YLast-YFirst)/Flt(YL);
    K: Integer := V'First;
    P: VecFltVec;
    ---
    P1: VecFlt renames P(1);
    X1: Flt renames P1.X(1);
    Y1: Flt renames P1.X(2);
    Z1: Flt renames P1.Y;
    ---
    P2: VecFlt renames P(2);
    X2: Flt renames P2.X(1);
    Y2: Flt renames P2.X(2);
    Z2: Flt renames P2.Y;
    ---
    P3: VecFlt renames P(3);
    X3: Flt renames P3.X(1);
    Y3: Flt renames P3.X(2);
    Z3: Flt renames P3.Y;
  begin
    while K<V'Last loop
      X1 := V(K);
      Y1 := V(K+1);
      Z1 := V(K+2);
      ---
      X2 := X1+HalfDX;
      Y2 := Y1;
      Z2 := Empty;
      ---
      X3 := X1;
      Y3 := Y1+HalfDY;
      Z3 := Empty;
      ---
      Improve_Min(P,Eps);
      V(K)   := X1;
      V(K+1) := Y1;
      V(K+2) := Z1;
      K := K+3;
    end loop;
  end Improve_Min;

  function Local_Minima(Z: Values; Dist: Positive; Eps: in Radius) return Flt_Vec is
    V: Flt_Vec := Rough_Local_Minima(Z,Dist);
  begin
    if V'Length>0 then Improve_Min(V,Eps); end if;
    return V;
  end Local_Minima;

  function Local_Minima(Dist: Positive; Eps: in Radius) return Flt_Vec is
    V: Flt_Vec := Rough_Local_Minima(Dist);
  begin
    if V'Length>0 then Improve_Min(V,Eps); end if;
    return V;
  end Local_Minima;

  procedure NegVal(V: in out Flt_Vec) is
    K: Integer := V'First;
  begin
    while K<V'Last loop
      K := K+2;
      V(K) := -V(K);
      K := K+1;
    end loop;
  end NegVal;

  procedure Show_List(N: in String; V: in Flt_Vec) is
    I: Integer := 1;
    K: Integer := V'First;
  begin
    if V'Length=0 then
      Show0(N & "  0   empty list");
    else
      while K<V'Last loop
        Show1(N & Strng(I,3) & " ",V(K),False);
        Show2(" ",V(K+1),V(K+2));
        I := I+1;
        K := K+3;
      end loop;
    end if;
  end Show_List;

  procedure Write(FileName: in String; V: in Flt_Vec; Decimal: in Boolean := False) is
    package Int_IO renames Ada.Integer_Text_IO;
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Writing " & FileName); end if;
    Create(F,Out_File,FileName);
    Int_IO.Put(F,V'Length);
    New_Line(F);
    for I in V'Range loop
      Put(F,V(I),Decimal);
    end loop;
    Close(F);
  end Write;

end Flts.Minima;
