with Globals, Strings, Protected_Counters;
use Globals, Strings, Protected_Counters;

pragma Elaborate_All (Globals,Strings,Protected_Counters);

package body N_M is

  procedure Copy(P1: in VecFlt; P2: in out VecFlt) is
    X1: Vec renames P1.X;
    X2: Vec renames P2.X;
  begin
    for I in VRange loop Copy(X1(I),X2(I)); end loop;
    P2.Y := P1.Y;
  end Copy;

  procedure Swap(P1,P2: in out VecFlt) is
    Y1: constant Flt := P1.Y;
    X1: Vec renames P1.X;
    X2: Vec renames P2.X;
  begin
    for I in VRange loop Swap(X1(I),X2(I)); end loop;
    P1.Y := P2.Y;
    P2.Y := Y1;
  end Swap;

  procedure Sort(P: in out VecFltVec) is
    Done: Boolean;
  begin
    loop
      Done := True;
      for J in reverse VRange loop
        if P(J).Y>P(J+1).Y then Swap(P(J),P(J+1)); Done := False; end if;
      end loop;
      exit when Done;
    end loop;
  end Sort;

  procedure Centroid(P: in VecFltVec; P0: in out VecFlt) is
    --- center of P(1),..,P(Dim)
    Fac: constant Flt := One/Flt(Dim);
    X0: Vec renames P0.X;
  begin
    for I in VRange loop
      Sum(P(1).X(I),P(2).X(I),X0(I));
      for J in 3 .. Dim loop
        Add(P(J).X(I),X0(I));
      end loop;
      Mult(Fac,X0(I));
    end loop;
    P0.Y := Empty;
  end Centroid;

  procedure Shrink(P: in out VecFltVec) is
    --- move P(2),..,P(Dim1) closer to P(1)
    X1: Vec renames P(1).X;
  begin
    for J in 2 .. Dim1 loop
      declare
        XJ: Vec renames P(J).X;
      begin
        for I in VRange loop
          Sub(X1(I),XJ(I));
          Mult(Sigma,XJ(I));
          Add(X1(I),XJ(I));
        end loop;
      end;
      P(J).Y := Empty;
    end loop;
  end Shrink;

  function NumEmpty(P: VecFltVec) return Natural is
    N: Integer := 0;
  begin
    for J in VRange1 loop
      if P(J).Y=Empty then N := N+1; end if;
    end loop;
    return N;
  end NumEmpty;

  procedure Fill(P: in out VecFltVec) is
    --- compute Yj=F(Xj)
    PC: Protected_Counter(1);
    procedure Fill_Loop is
      J: Positive;
    begin
      loop
        PC.Next(J);
        exit when J>Dim1;
        P(J).Y := F(P(J).X);
      end loop;
    end Fill_Loop;
    NT: constant Natural := Reserve_Tasks(Integer'Max(NumEmpty(P)-1,0),Fill_Parallel);
  begin
    for J in VRange1 loop
      for I in VRange loop Center(P(J).X(I)); end loop;
    end loop;
    if NT=0 then
      Fill_Loop;
    else
      declare
        Task_Error: Boolean := False;
        task type Fill_Task_Type is end Fill_Task_Type;
        task body Fill_Task_Type is
        begin
          Proper_Rounding;
          Fill_Loop;
        exception
          when others => Task_Error := True; raise;
        end Fill_Task_Type;
        ---
        Fill_Task: array(1 .. NT) of Fill_Task_Type;
        pragma Warnings (Off,Fill_Task);
      begin
        Fill_Loop;
        if Task_Error then raise Task_Exception with "Task_Error"; end if;
      end;
      Free_Tasks(NT);
    end if;
  end Fill;

  procedure FindMin(P: in out VecFltVec) is
    procedure Fill_Values is new Fill (F => F);
    P1: VecFlt renames P(1);
    Y1: Flt renames P1.Y;
    PDim1: VecFlt renames P(Dim1);
    XDim1: Vec renames PDim1.X;
    YDim1: Flt renames PDim1.Y;
    ---
    YMin: Flt := Empty;
    P0,PR: VecFlt;
    PC: VecFlt renames PR;
    X0: Vec renames P0.X;
    XR: Vec renames PR.X;
    XC: Vec renames PC.X;
  begin
    loop
      Fill_Values(P);
      Sort(P);
      if Y1<YMin then
        if Done(P) then return; end if;
        YMin := Y1;
      end if;
      Centroid(P,P0);
      for I in VRange loop
        Diff(X0(I),XDim1(I),XR(I));
        Mult(Alpha,XR(I));
        Add(X0(I),XR(I));
      end loop;
      PR.Y := F(XR);                 -- PR := reflected(PDim1)
      if PR.Y<P(Dim).Y then
        if PR.Y >= Y1 then
          Copy(PR,PDim1);
          if Verbose then
            Show0("nm: reflected");
          end if;
        else                         -- PR.Y<P1.Y
          for I in VRange loop
            Diff(XR(I),X0(I),XDim1(I));
            Mult(Gamma,XDim1(I));
            Add(X0(I),XDim1(I));
          end loop;
          YDim1 := F(XDim1);         -- PDim1 := expanded(PDim1)
          if YDim1 >= PR.Y then
            Copy(PR,PDim1);
            if Verbose then
              Show0("nm: reflected");
            end if;
          elsif Verbose then
            Show0("nm: expanded");
          end if;
        end if;
      else                           -- PR.Y >= P(Dim).Y
        for I in VRange loop
          Diff(XDim1(I),X0(I),XC(I));
          Mult(Rho,XC(I));
          Add(X0(I),XC(I));
        end loop;
        PC.Y := F(XC);               -- PC := contracted(PDim1)
        if PC.Y<YDim1 then
          Copy(PC,PDim1);
          if Verbose then
            Show0("nm: contracted");
          end if;
        else
          Shrink(P);
          if Verbose then
            Show0("nm: did shrink");
          end if;
        end if;
      end if;
    end loop;
  end FindMin;

end N_M;
