with Ada.Command_Line, Globals, Strings;

pragma Elaborate_All (Ada.Command_Line,Strings,Globals);

package body Rationals is

  function Sqr(I: Lint) return LInt is
    pragma Unsuppress (Overflow_Check);
  begin
    return I*I;
  end Sqr;

  function GCD(N,D: LInt) return LInt is
    use Globals;
    AN: LInt := abs(N);
    AD: LInt := abs(D);
    Tmp: LInt;
  begin
    while (AN /= 0) loop
      Tmp  := AN;
      AN := AD mod Tmp;
      AD := Tmp;
    end loop;
    if AD=0 then raise CONSTRAINT_ERROR; end if;
    return AD;
  end GCD;

  function Width(I: LInt; Base: Positive := 10) return Positive is
    B: constant LInt := LInt(Base);
    L: Integer;
    K: LInt;
  begin
    K := abs(I);
    if I<0 then L := 2; else L := 1; end if;
    while K >= B loop
      K := K/B;
      L := L+1;
    end loop;
    return L;
  end Width;

  function Strng(I: LInt; W: Positive) return String is
    S: String(1 .. W);
  begin
    LInt_IO.Put(S,I);
    return S;
  end Strng;

  function Strng(I: LInt) return String is
    S: String(1 .. Width(I));
  begin
    LInt_IO.Put(S,I);
    return S;
  end Strng;

  function Value(S: String) return LInt is
    pragma Unsuppress (Overflow_Check);
    use Strings;
    Neg: Boolean := False;
    I: LInt := 0;
  begin
    for K in S'Range loop
      if S(K)='-' then
        Neg := True;
      elsif S(K) /= ' ' and then S(K) /= '+' then
        I := 10*I+LInt(Val(S(K),9));
      end if;
    end loop;
    if Neg then I := -I; end if;
    return I;
  end Value;

  procedure Show1(N: in String; I: in LInt; NewLine: in Boolean := True) is
    use Strings, Globals;
    P: constant File_Access := Default_Output.all;
  begin
    Put(P,N);
    LInt_IO.Put(P.all,I,Width(I));
    if NewLine then New_Line(P); end if;
  end Show1;

  function Prompt(N: String) return LInt is
    use Strings;
    S: constant String := Prompt(N);
  begin
    return Value(S);
  end Prompt;

  procedure Put(F: in File_Type; I: in LInt; Dummy: in Boolean := False) is
    use Strings;
    pragma Unreferenced(Dummy);
  begin
    LInt_IO.Put(F,I,Width(I)+1);
    New_Line(F);
  end Put;

  procedure Get(F: in File_Type; I: in out LInt; Dummy: in Boolean := False) is
    pragma Unreferenced(Dummy);
  begin
    LInt_IO.Get(F,I);
  end Get;

  --------------------------

  ------------------------------

  function "-"(Q: Rational) return Rational is
  begin
    return (-Q.Num,Q.Den);
  end "-";

  function "+"(P,Q: Rational) return Rational is
    pragma Unsuppress (Overflow_Check);
    G: constant LInt := GCD(P.Den,Q.Den);
    PD: constant Lint := P.Den/G;
    QD: constant Lint := Q.Den/G;
  begin
    return (P.Num*QD+PD*Q.Num)/(PD*Q.Den);
  end "+";

  function "-"(P,Q: Rational) return Rational is
    pragma Unsuppress (Overflow_Check);
    G: constant LInt := GCD(P.Den,Q.Den);
    PD: constant Lint := P.Den/G;
    QD: constant Lint := Q.Den/G;
  begin
    return (P.Num*QD-PD*Q.Num)/(PD*Q.Den);
  end "-";

  function "*"(P,Q: Rational) return Rational is
    pragma Unsuppress (Overflow_Check);
    A: constant Rational := P.Num/Q.Den;
    B: constant Rational := Q.Num/P.Den;
  begin
    return (A.Num*B.Num,A.Den*B.Den);
  end "*";

  function "/" (P,Q: Rational) return Rational is
    pragma Unsuppress (Overflow_Check);
    A: constant Rational := P.Num/Q.Num;
    B: constant Rational := Q.Den/P.Den;
  begin
    return (A.Num*B.Num,A.Den*B.Den);
  end "/";

  function "mod"(P,Q: Rational) return Rational is
    G: constant LInt := GCD(P.Den,Q.Den);
  begin
    return (P.Num*(Q.Den/G) mod Q.Num*(P.Den/G))/(P.Den*(Q.Den/G));
  end "mod";

  function Inv(Q: Rational) return Rational is
  begin
    if Q.Num>0 then return (Q.Den,Q.Num); end if;
    return (-Q.Den,-Q.Num);
  end Inv;

  function Sqr(Q: Rational) return Rational is
    pragma Unsuppress (Overflow_Check);
  begin
    return (Q.Num*Q.Num,Q.Den*Q.Den);
  end Sqr;

  function IsZero(Q: Rational) return Boolean is
  begin
    return Q.Num=0;
  end IsZero;

  function Sign(Q: Rational) return Integer is
  begin
    if Q.Num>0 then return 1; end if;
    if Q.Num<0 then return -1; end if;
    return 0;
  end Sign;

  function "abs"(Q: Rational) return Rational is
  begin
    return (abs(Q.Num),Q.Den);
  end "abs";

  function "<"(P,Q: Rational) return Boolean is
    pragma Unsuppress (Overflow_Check);
    G: constant LInt := GCD(P.Den,Q.Den);
    PD: constant Lint := P.Den/G;
    QD: constant Lint := Q.Den/G;
  begin
    return (P.Num*QD)<(Q.Num*PD);
  end "<";

  ----------------------- L

  function LRat(N: Lint; D: LInt := 1) return Rational is
    G: constant LInt := GCD(N,D);
  begin
    if D>0 then
      return (N/G,D/G);
    end if;
    if D=0 then raise CONSTRAINT_ERROR; end if;
    return ((-N)/G,(-D)/G);
  end LRat;

  function LNum(Q: Rational) return Lint is
  begin
    return Q.Num;
  end LNum;

  function LDen(Q: Rational) return Lint is
  begin
    return Q.Den;
  end LDen;

  procedure LFloorPlus(Q: in out Rational; C: out LInt) is
    --- continued fractions step
  begin
    C := Q.Num/Q.Den;     -- rounds toward 0
    if C*Q.Den=Q.Num then
      Q := Rat_Zero;
    else
      if Q.Num<0 then C := C-1; end if;
      Q := Q.Den/(Q.Num-C*Q.Den);
    end if;
  end LFloorPlus;

  function LFloor(Q: Rational) return LInt is
    I: constant LInt := Q.Num/Q.Den; -- rounds toward 0
  begin
    if Q.Num >= 0 then return I; end if;
    if I*Q.Den=Q.Num then return I; end if;
    return I-1;
  end LFloor;

  function LCeiling(Q: Rational) return LInt is
    I: constant LInt := Q.Num/Q.Den; -- rounds toward 0
  begin
    if Q.Num <= 0 then return I; end if;
    if I*Q.Den=Q.Num then return I; end if;
    return I+1;
  end LCeiling;

  function "mod"(P: Rational; I: LInt) return Rational is
  begin
    return (P.Num mod (I*P.Den),P.Den);
  end "mod";

  function LSum(I: LInt; Q: Rational) return Rational is
    pragma Unsuppress (Overflow_Check);
  begin
    return (I*Q.Den+Q.Num,Q.Den);
  end LSum;

  function LSum(Q: Rational; I: LInt) return Rational is
    pragma Unsuppress (Overflow_Check);
  begin
    return (Q.Num+I*Q.Den,Q.Den);
  end LSum;

  function LDiff(I: LInt; Q: Rational) return Rational is
    pragma Unsuppress (Overflow_Check);
  begin
    return (I*Q.Den-Q.Num,Q.Den);
  end LDiff;

  function LDiff(Q: Rational; I: LInt) return Rational is
    pragma Unsuppress (Overflow_Check);
  begin
    return (Q.Num-I*Q.Den,Q.Den);
  end LDiff;

  function LProd(I: LInt; Q: Rational) return Rational is
    pragma Unsuppress (Overflow_Check);
  begin
    return (I*Q.Num)/Q.Den;
  end LProd;

  function LProd(Q: Rational; I: LInt) return Rational is
    pragma Unsuppress (Overflow_Check);
  begin
    return (Q.Num*I)/Q.Den;
  end LProd;

  function LQuot(I: LInt; Q: Rational) return Rational is
    pragma Unsuppress (Overflow_Check);
  begin
    return (I*Q.Den)/Q.Num;
  end LQuot;

  function LQuot(Q: Rational; I: LInt) return Rational is
    pragma Unsuppress (Overflow_Check);
  begin
    return Q.Num/(Q.Den*I);
  end LQuot;

  ----------------------- I

  function IRat(N: Integer; D: Integer := 1) return Rational is
  begin
    return (Lint(N)/Lint(D));
  end IRat;

  function INum(Q: Rational) return Integer is
    use LInt_IO;
  begin
    return Integer(Q.Num);
  exception
    when CONSTRAINT_ERROR =>
      LInt_IO.Put(Q.Num,Width(Q.Num));
      New_Line;
      raise;
  end INum;

  function IDen(Q: Rational) return Integer is
    use LInt_IO;
  begin
    return Integer(Q.Den);
  exception
    when CONSTRAINT_ERROR =>
      LInt_IO.Put(Q.Den,Width(Q.Den));
      New_Line;
      raise;
  end IDen;

  function IFloor(Q: Rational) return Integer is
    pragma Unsuppress (Overflow_Check);
  begin
    return Integer(LFloor(Q));
  end IFloor;

  function ICeiling(Q: Rational) return Integer is
    pragma Unsuppress (Overflow_Check);
  begin
    return Integer(LCeiling(Q));
  end ICeiling;

  function "mod"(P: Rational; I: Integer) return Rational is
  begin
    return (P.Num mod (Lint(I)*P.Den),P.Den);
  end "mod";

  function ISum(I: Integer; Q: Rational) return Rational is
    pragma Unsuppress (Overflow_Check);
  begin
    return (LInt(I)*Q.Den+Q.Num,Q.Den);
  end ISum;

  function ISum(Q: Rational; I: Integer) return Rational is
    pragma Unsuppress (Overflow_Check);
  begin
    return (Q.Num+LInt(I)*Q.Den,Q.Den);
  end ISum;

  function IDiff(I: Integer; Q: Rational) return Rational is
    pragma Unsuppress (Overflow_Check);
  begin
    return (LInt(I)*Q.Den-Q.Num,Q.Den);
  end IDiff;

  function IDiff(Q: Rational; I: Integer) return Rational is
    pragma Unsuppress (Overflow_Check);
  begin
    return (Q.Num-LInt(I)*Q.Den,Q.Den);
  end IDiff;

  function IProd(I: Integer; Q: Rational) return Rational is
    pragma Unsuppress (Overflow_Check);
  begin
    return (LInt(I)*Q.Num)/Q.Den;
  end IProd;

  function IProd(Q: Rational; I: Integer) return Rational is
    pragma Unsuppress (Overflow_Check);
  begin
    return (Q.Num*LInt(I))/Q.Den;
  end IProd;

  function IPower(Q: Rational; I: Integer) return Rational is
    pragma Unsuppress (Overflow_Check);
    --- 0**0=1
  begin
    if I>0 then return (Q.Num**I,Q.Den**I); end if;
    if I<0 then return IPower(Inv(Q),-I); end if;
    return Rat_One;
  end IPower;

  function IQuot(I: Integer; Q: Rational) return Rational is
    pragma Unsuppress (Overflow_Check);
  begin
    return (LInt(I)*Q.Den)/Q.Num;
  end IQuot;

  function IQuot(Q: Rational; I: Integer) return Rational is
    pragma Unsuppress (Overflow_Check);
  begin
    return Q.Num/(Q.Den*LInt(I));
  end IQuot;

  -----------------------------

  function Strng(Q: Rational) return String is
  begin
    return Strng(Q.Num) & "/" & Strng(Q.Den);
  end Strng;

  function Value(S: String) return Rational is
    Num: LInt;
    Den: LNat;
  begin
    for K in S'First+1 .. S'Last-1 loop
      if S(K)='/' then
        Num := Value(S(S'First .. K-1));
        Den := Value(S(K+1 .. S'Last));
        return LRat(Num,Den);
      end if;
    end loop;
    Num := Value(S);
    return (Num,1);
  end Value;

  procedure Normalize(Q: in out Rational) is
    G: constant LInt := GCD(Q.Num,Q.Den);
  begin
    if G /= 1 then
      Q.Num := Q.Num/G;
      Q.Den := Q.Den/G;
    end if;
  end Normalize;

  function Normalize(Q: Rational) return Rational is
    G: constant LInt := GCD(Q.Num,Q.Den);
  begin
    if G /= 1 then
      Show1("Warning: found un-normalized rational ",Q);
    end if;
    return (Q.Num/G,Q.Den/G);
  end Normalize;

  procedure Show1(N: in String; Q: in Rational; NewLine: in Boolean := True) is
    use LInt_IO;
    P: constant Rational := Normalize(Q);
  begin
    Put(N);
    LInt_IO.Put(P.Num,Width(P.Num));
    Put("/");
    LInt_IO.Put(P.Den,Width(P.Den));
    if NewLine then New_Line; end if;
  end Show1;

  function Prompt(N: String) return Rational is
    use Strings;
    S: constant String := Prompt(N);
  begin
    return Value(S);
  end Prompt;

  function GetArg return Rational is
    use Globals;
    S: constant String := Ada.Command_Line.Argument(ArgIndex);
  begin
    ArgIndex := ArgIndex+1;
    return Value(S);
  end GetArg;

  procedure Put(F: in File_Type; Q: in Rational; Dummy: in Boolean := False) is
    use Strings;
    pragma Unreferenced(Dummy);
  begin
    Put(F,Q.Num);
    Put(F,Q.Den);
  end Put;

  procedure Get(F: in File_Type; Q: in out Rational; Dummy: in Boolean := False) is
    pragma Unreferenced(Dummy);
  begin
    Get(F,Q.Num);
    Get(F,Q.Den);
  end Get;

  ----------------------

  function EvenOverOdd(Q: Rational) return Boolean is
    P: constant Rational := Normalize(Q);
  begin
    return ((P.Num mod 2)=0) and then ((P.Den mod 2)=1);
  end EvenOverOdd;

  function OddOverEven(Q: Rational) return Boolean is
    P: constant Rational := Normalize(Q);
  begin
    return ((P.Num mod 2)=1) and then ((P.Den mod 2)=0);
  end OddOverEven;

  function OddOverOdd(Q: Rational) return Boolean is
    P: constant Rational := Normalize(Q);
  begin
    return ((P.Num mod 2)=1) and then ((P.Den mod 2)=1);
  end OddOverOdd;

  function ContFrac_Length(Q: Rational) return Natural is
    L: Natural := 0;
    C: Lint;
    P: Rational := Q;
  begin
    while P.Num /= 0 loop
      LFloorPlus(P,C);
      L := L+1;
    end loop;
    return L;
  end ContFrac_Length;

  function ContFrac(Q: Rational) return Int_Vec is
    LMax: constant Integer := 127;
    K: Integer := -1;
    P: Rational := Q;
    C: Int_Vec(0 .. LMax);
  begin
    while P.Num /= 0 loop
      K := K+1;
      LFloorPlus(P,LInt(C(K)));
    end loop;
    return C(0 .. K);
  end ContFrac;

  procedure PnQn(C: in Int_Vec; Pn0,Pn1,Qn0,Qn1: out LInt) is
    use Globals;
    -- Pn := c_n*Pn0+Pn1
    -- Pn := c_n*Pn0+Pn1
    -- assuming C is non-empty
    F: constant Integer := C'First;
    Cn,Pn2,Qn2,Det: LInt;
  begin
    Pn0 := LInt(C(F));
    Pn1 := 1;
    Qn0 := 1;
    Qn1 := 0;
    for N in F+1 .. C'Last loop
      Cn := LInt(C(N));
      if Cn <= 0 then raise Sorry with "non-positive c_n"; end if;
      Pn2 := Pn1;
      Pn1 := Pn0;
      Pn0 := Cn*Pn1+Pn2;
      ---
      Qn2 := Qn1;
      Qn1 := Qn0;
      Qn0 := Cn*Qn1+Qn2;
    end loop;
    Det := Pn0*Qn1-Qn0*Pn1;
    if Det /= (-1)**(C'Length mod 2) then raise Sorry; end if;
  end PnQn;

  function PnQn(C: Int_Vec) return IMat2 is
    Pn0,Pn1,Qn0,Qn1: Lint;
    M: IMat2;
  begin
    PnQn(C,Pn0,Pn1,Qn0,Qn1);
    M(1,1) := Integer(Pn0);
    M(1,2) := Integer(Pn1);
    M(2,1) := Integer(Qn0);
    M(2,2) := Integer(Qn1);
    return M;
  end PnQn;

  function PnQn(Q: Rational) return IMat2 is
    use Globals;
    C: constant Int_Vec := ContFrac(Q);
    Pn0,Pn1,Qn0,Qn1: Lint;
    M: IMat2;
  begin
    PnQn(C,Pn0,Pn1,Qn0,Qn1);
    if Pn0 /= Q.Num then raise Sorry; end if;
    if Qn0 /= Q.Den then raise Sorry; end if;
    M(1,1) := Integer(Pn0);
    M(1,2) := Integer(Pn1);
    M(2,1) := Integer(Qn0);
    M(2,2) := Integer(Qn1);
    return M;
  end PnQn;

end Rationals;
